/**********************************************************************
 
	Copyright (C) 2009 Hirohisa MORI <joshua45@globalbase.org>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/


#include	"utils.h"
#include	"netutils.h"
#include	"memory_debug.h"

typedef struct url_item_list {
	struct url_item_list *	next;
	L_CHAR *		str;
} URL_ITEM_LIST;


URL_ITEM_LIST *
get_uil_1(L_CHAR * str,char * pre,char * noitem)
{
URL_ITEM_LIST * ret, * r;
	if ( str ) {
		ret = d_alloc(sizeof(*ret));
		ret->str = ll_copy_str(str);
		ret->next = 0;
		if ( pre ) {
			r = get_uil_1(l_string(std_cm,pre),0,0);
			r->next = ret;
			ret = r;
		}
	}
	else if ( noitem ) {
		ret = d_alloc(sizeof(*ret));
		ret->str = nl_copy_str(std_cm,noitem);
		ret->next = 0;
	}
	else	ret = 0;
	return ret;
}

URL_ITEM_LIST *
get_uil_server(L_CHAR * str,char * pre,char * noitem)
{
URL_ITEM_LIST * ret, * r;
L_CHAR * _str;
L_CHAR * p, * q;
int edf;
	if ( str ) {
		_str = ll_copy_str(str);
		p = q = _str;
		ret = 0;
		for ( ; ; ) {
			for ( q = p ; *q && *q != '.' ; q ++ );
			if ( *q == 0 )
				edf = 1;
			else	edf = 0;
			*q = 0;
			r = get_uil_1(p,0,0);
			r->next = ret;
			ret = r;
			if ( edf == 1 )
				break;
			p = q+1;
		}
		if ( pre ) {
			r = get_uil_1(l_string(std_cm,pre),0,0);
			r->next = ret;
			ret = r;
		}
	}
	else if ( noitem ) {
		ret = get_uil_1(0,0,noitem);
	}
	else	ret = 0;
	return ret;
}


URL_ITEM_LIST *
get_uil_db(L_CHAR * str,char * pre,char * noitem)
{
URL_ITEM_LIST * ret, * r, * r1;
L_CHAR * _str;
L_CHAR * p, * q;
int edf;
	if ( str ) {
		_str = ll_copy_str(str);
		p = q = _str;
		ret = 0;
		for ( ; *p == '/' ; p ++ );
		for ( ; ; ) {
			for ( q = p ; *q && *q != '/' ; q ++ );
			if ( *q == 0 )
				edf = 1;
			else	edf = 0;
			*q = 0;
			r = get_uil_1(p,0,0);
			r->next = ret;
			ret = r;
			if ( edf == 1 )
				break;
			p = q+1;
			for ( ; *p == '/' ; p ++ );
			if ( *p == 0 )
				break;
		}
		r = ret;
		ret = 0;
		for ( ; r ; ) {
			r1 = r;
			r = r->next;
			r1->next = ret;
			ret = r1;
		}
		if ( pre ) {
			r = get_uil_1(l_string(std_cm,pre),0,0);
			r->next = ret;
			ret = r;
		}
	}
	else if ( noitem ) {
		ret = get_uil_1(0,0,noitem);
	}
	else	ret = 0;
	return ret;
}


URL_ITEM_LIST *
get_url_item_list(URL * a)
{
URL_ITEM_LIST * ret;
URL_ITEM_LIST ** p;
char buffer[10];
	ret = get_uil_1(a->proto,0,"xlp");
	p = &ret->next;
	*p = get_uil_server(a->server,0,"localhost");
	for ( p = &ret ; *p ; p = &(*p)->next);
	sprintf(buffer,"%i",a->port);
	*p = get_uil_1(l_string(std_cm,buffer),0,0);
	for ( p = &ret ; *p ; p = &(*p)->next);
	*p = get_uil_1(a->agent,0,"@");
	for ( p = &ret ; *p ; p = &(*p)->next);
	*p = get_uil_db(a->db,0,0);
	for ( p = &ret ; *p ; p = &(*p)->next);
	*p = get_uil_1(a->resource,0,"-");
	for ( p = &ret ; *p ; p = &(*p)->next);
	*p = get_uil_1(a->query,"?",0);
	for ( p = &ret ; *p ; p = &(*p)->next);
	*p = get_uil_1(a->name,"#",0);
	return ret;
}

void
free_url_item_list(URL_ITEM_LIST * a)
{
URL_ITEM_LIST * b;
	for ( ; a ; ) {
		b = a->next;
		d_f_ree(a->str);
		d_f_ree(a);
		a = b;
	}
}

URL_DISTANCE_T
url_distance(URL * a,URL * b)
{
URL_ITEM_LIST * a1,* b1;
URL_ITEM_LIST * a2, * b2;
URL_DISTANCE_T ret;
	a1 = get_url_item_list(a);
	b1 = get_url_item_list(b);
	
	ret.c = 0;
	for ( a2 = a1 , b2 = b1 ; a2 && b2 && l_strcmp(a2->str,b2->str) == 0  ; a2 = a2->next , b2 = b2->next , ret.c ++ );
	ret.d = 0;
	for ( ; a2 ; a2 = a2->next , ret.d ++ );
	for ( ; b2 ; b2 = b2->next , ret.d ++ );
	free_url_item_list(a1);
	free_url_item_list(b1);
	return ret;
}


